#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
from itertools import chain
from typing import Dict, List


from .interfaces import Table, RowWriter


class PostgresCopyRowWriter(RowWriter):
    """
    Write rows in the PostgreSQL COPY TABLE format. This is the most efficient
    way of loading data into Postgres and it's the way used by the pg_dump
    utility for creating backups.
    """

    def write_rows(self, table: "Table", rows: List[Dict]):
        columns = rows[0].keys()
        print(f"COPY {table.name} ({', '.join(columns)}) FROM stdin;")
        for row in rows:
            print("\t".join([self.copy_format(row[col]) for col in columns]))
        print("\.")

    @staticmethod
    def copy_format(value):
        if value is None:
            return "\\N"
        elif isinstance(value, str):
            return value
        elif isinstance(value, int) or isinstance(value, float):
            return str(value)
        elif isinstance(value, datetime.date):
            return value.isoformat()
        else:
            raise TypeError(f"Unable to figure out what to do with {value} of type {type(value)}")


class DictionaryKeeper(RowWriter):
    """
    Retain rows as a list of dictionaries under each table name. Intended to support
    reingestion.

    Ask for the property ``items`` to read the items afterwards.
    """

    def __init__(self):
        self.items = []

    def write_rows(self, table: Table, rows: List[Dict]):
        self.items.append({"table": table.name, "items": rows})

    def close(self):
        """A no-op; nothing to do here"""


class UniquifyingRowWriter(RowWriter):
    """
    Ensure that only a single instance of each row gets dumped out.
    """

    def __init__(self, underlying: RowWriter):
        self.underlying = underlying
        self.seen = {}

    def write_rows(self, table: "Table", rows: List[Dict]):
        seen = self.seen.setdefault(table.name, [])
        new_rows = [row for row in rows if row not in seen]
        if len(new_rows) > 0:
            self.underlying.write_rows(table, new_rows)
            self.seen[table.name] += new_rows

    def close(self):
        self.underlying.close()


class TopologicalSortFailed(Exception):
    pass


class TopologicallySortingRowWriter(RowWriter):
    """
    Topologically sorts the tables before outputting their rows.

    Because you can have an arbitrary graph topology of foreign keys in your database, you have to worry
    about what order you try to input rows when you read them. If you try to read a row before you have
    loaded a row that it depends on, the load will fail. For instance, science product locators point to
    projects, but execblocks point to both projects and science product locators. By starting from the
    project, the program will want to output execblocks first due to alphabetical order. This would be
    difficult to do in a query so instead we do it here because we have all the objects we need to figure
    out what order things should go in.

    The algorithm that gives you this ordering across a digraph is called topological sort. This implementation
    was created while reading notes in the slides at:
    https://courses.cs.washington.edu/courses/cse326/03wi/lectures/RaoLect20.pdf
    """

    def __init__(self, underlying: RowWriter):
        self.tables_rows: Dict["Table", List[Dict]] = {}
        self.underlying = underlying

    def write_rows(self, table: "Table", rows: List[Dict]):
        # We've been asked to output some rows for this table.
        # What we must now do is keep track of this table and the in-bound relationships with it.
        # Knowing which tables we need to output will give us the vertices of the graph we need to sort.
        self.tables_rows[table] = rows

    def close(self):
        # Now that we know all the tables we're going to be asked to output, we can compute their edges.
        # We have an edge for each relationship from that table to another table we've been asked to output
        vertices = self.tables_rows.keys()

        # There is an object called graphlib.TopologicalSorter in Python 3.9+ that can replace
        # the code below.

        # Now we can get the exact set of relations we actually care about by getting all
        # of the relations between all of the tables and filtering out relations that pertain
        # to tables we do not have data for. We also ignore self-joins here because they increase
        # the in-degree by one but cannot ever be decreased, so we avoid an infinite loop by
        # ignoring them. We can assume that, to whatever extent this program handles self-joins,
        # it will obtain upper level rows before their dependencies programmatically, so we don't
        # have to handle it manually here.
        all_relations = chain.from_iterable(v.relations() for v in vertices)
        relations = [
            r
            for r in all_relations
            if r.source_table in vertices and r.destination_table in vertices and r.source_table != r.destination_table
        ]

        # we can now compute the in-degree values for each vertex by computing how many relationships
        # from other tables we care about point to this table
        in_degrees = {}
        for vertex_table in vertices:
            in_degrees[vertex_table] = sum(1 for relation in relations if relation.destination_table == vertex_table)

        # now that we know the in_degrees we can proceed with the sort
        remaining = list(vertices)
        while len(remaining) > 0:
            # find a vertex in remaining whose in-degrees is 0
            table = None
            for table in remaining:
                if in_degrees[table] == 0:
                    break

            # if we didn't find one, throw an exception
            if table is None:
                raise TopologicalSortFailed("Unable to locate a table with no references!")

            # pass along the rows we found
            self.underlying.write_rows(table, self.tables_rows[table])

            # now we can remove this vertex from the remaining vertices
            # and reduce its neighbor's input degree by one
            remaining.remove(table)
            for relation in relations:
                if relation.source_table == table and relation.destination_table in in_degrees:
                    in_degrees[relation.destination_table] -= 1

        # if we made it here, we ran out of remaining tables and we can conclude the process
        self.underlying.close()
